from django.conf import settings
from django.contrib.auth import login
from django.db import transaction

from . import exceptions, models


def app_register(request):
    try:
        with transaction.atomic():

            user = settings.AUTH_USER_MODEL.objects.create_user(
                username=request.GET["key"], password=request.GET["key"]
            )
            user.lnauthkey.linking_key = request.GET["key"]
            user.lnauthkey.save()
            user.save()
    except models.LnAuthKey.IntegrityError:
        raise exceptions.DjangoAuthException("User already registered.")

    login(request, user)


def app_login(request):
    try:
        user = settings.AUTH_USER_MODEL.objects.get(
            lnauthkey__linking_key=request.GET["key"]
        )
    except settings.AUTH_USER_MODEL.DoesNotExist:
        raise exceptions.DjangoAuthException("User not registered.")

    login(request, user)



import binascii
import os

import bech32
from django.conf import settings
from django.core.cache import cache
from django.urls import reverse
from secp256k1 import PublicKey

from . import exceptions


def get_auth_url(action: str):
    k1 = os.urandom(32)
    reverse_url = reverse("lnauth_django:ln_auth_url")

    cache.set(
        f"lnauth-k1-{k1.hex()}",
        action,
        timeout=getattr(settings, "LNURL_AUTH_K1_TIMEOUT", 60 * 60),
    )

    url = f"{settings.LNURL_AUTH_ROOT_DOMAIN}{reverse_url}?tag=login&k1={k1.hex()}&action={action}"

    bech32_url = bech32.bech32_encode(
        "lnurl",
        bech32.convertbits(url.encode("utf-8"), 8, 5),
    )

    return bech32_url


def verify_ln_auth(k1: str, sig: str, linking_key: str):

    try:
        k1_bytes = binascii.unhexlify(k1)
        sig_bytes = binascii.unhexlify(sig)
        linking_key_bytes = binascii.unhexlify(linking_key)
    except binascii.Error:
        raise exceptions.LnAuthException("Invalid hex.")

    if not cache.delete(f"lnauth-k1-{k1_bytes.hex()}"):
        raise exceptions.LnAuthException("K1 does not exist.")

    linking_key_pubkey = PublicKey(linking_key_bytes, raw=True)

    sig_raw = linking_key_pubkey.ecdsa_deserialize(sig_bytes)

    if not linking_key_pubkey.ecdsa_verify(k1_bytes, sig_raw, raw=True):
        raise exceptions.InvalidSigException("Invalid signature.")



from django.urls import path

from .views import AuthURLProviderView, AuthURLView

app_name = "lnauth_django"
urlpatterns = [
    path(
        "ln-auth-get-url/", AuthURLProviderView.as_view(), name="ln_auth_url_provider"
    ),
    path("ln-auth/", AuthURLView.as_view(), name="ln_auth_url"),
]



from django.conf import settings
from django.db import models


class LnAuthKey(models.Model):
    user = models.OneToOneField(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    linking_key = models.CharField(max_length=255, null=True, blank=True, unique=True)



import binascii

import bech32
import django
from coincurve import PrivateKey
from django.conf import settings
from django.core.cache import cache
from django.urls import include, path

from .. import lnauth

settings.configure(
    DEBUG=True,
    ROOT_URLCONF=__name__,
    LNURL_AUTH_ROOT_DOMAIN="localhost:8000",
    INSTALLED_APPS=[
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "lnauth_django",
    ],
)

# Reconfigure Django's global settings object
django.setup()


urlpatterns = [
    path("lnauth_django/", include("lnauth_django.urls", namespace="lnauth_django")),
]


def test_ln_auth_flow():

    # Step 1: Generate a new private key and corresponding public key
    private_key = PrivateKey()
    public_key = private_key.public_key.format(compressed=False).hex()

    # Step 2: Get auth URL
    action = "test"
    auth_url_bech32 = lnauth.get_auth_url(action)

    # Step 3: Extract the k1 challenge from the URL
    _, auth_url_bytes = bech32.bech32_decode(auth_url_bech32)
    auth_url = bytes(bech32.convertbits(auth_url_bytes, 5, 8)).decode()
    k1 = auth_url.split("&")[1].split("=")[1]

    # Step 4: Sign the challenge with the private key
    signature = private_key.sign(binascii.unhexlify(k1), hasher=None)
    sig = signature.hex()

    # Step 5: Verify the signature
    lnauth.verify_ln_auth(k1, sig, public_key)

    # Cache should be empty at this point
    assert len(cache._cache) == 0



import logging

from django.http import JsonResponse
from django.views import generic

from . import django_auth, exceptions, lnauth

logger = logging.getLogger(__name__)


class AuthURLProviderView(generic.View):
    def get(self, request):
        if request.GET and "action" in request.GET:
            return JsonResponse(
                {"url": lnauth.get_auth_url(request.GET["action"])}, status=200
            )
        else:
            return JsonResponse({"message": "Invalid request."}, status=400)


class AuthURLView(generic.View):
    def get(self, request):
        if not request.GET:
            return JsonResponse({"message": "Invalid request."}, status=400)

        for key in ["tag", "k1", "action", "sig", "key"]:
            if key not in request.GET:
                return JsonResponse(
                    {"status": "ERROR", "reason": f"Invalid request. Missing '{key}'."},
                    status=400,
                )
        if request.GET["tag"] != "login":
            return JsonResponse(
                {"status": "ERROR", "reason": "Invalid request. Invalid tag."},
                status=400,
            )

        try:
            lnauth.verify_ln_auth(
                request.GET["k1"], request.GET["sig"], request.GET["key"]
            )
        except exceptions.LnAuthException as e:
            return JsonResponse(
                {"status": "ERROR", "reason": f"Invalid request. {e}"}, status=400
            )

        if request.GET["action"] == "login":

            try:
                django_auth.app_login(request)
            except exceptions.DjangoAuthException as e:
                return JsonResponse(
                    {"status": "ERROR", "reason": f"Unauthorized. {e}"}, status=401
                )

        elif request.GET["action"] == "register":
            try:
                django_auth.app_register(request)
            except exceptions.DjangoAuthException as e:
                return JsonResponse(
                    {"status": "ERROR", "reason": f"Unauthorized. {e}"}, status=401
                )
        else:
            return JsonResponse(
                {"status": "ERROR", "reason": "Invalid request. Invalid action."},
                status=400,
            )

        return JsonResponse({"status": "OK"}, status=200)
